import numpy as np
import cv2
import matplotlib.pyplot as plt

def map_screen_to_world(screen_params, cam_params, x_offset=0, y_offset=0, z_offset=0):
    """
    将屏幕坐标映射到世界坐标
    """
    # 确保平移向量是一维数组
    Tv2c = screen_params['screen_translation_vector'].reshape(-1)
    nc = cam_params['translation_vector'].reshape(-1)

    # print("Debug - Translation vectors:")
    # print(f"Screen translation vector: {Tv2c}")
    # print(f"Camera translation vector: {nc}")
    # print(f"Shapes: Tv2c {Tv2c.shape}, nc {nc.shape}")
    
    assert Tv2c.shape == nc.shape, f"Shape mismatch: Tv2c {Tv2c.shape} != nc {nc.shape}"
    
    # 提取旋转矩阵和平移向量
    #print('data_screen = ', screen_params)
    
    # Rv2c = data_screen[0]['screen_rotation_matrix']
    # Tv2c = data_screen[0]['screen_translation_vector']
    Rv2c = screen_params['screen_rotation_matrix']
    Tv2c = screen_params['screen_translation_vector']
    A = cam_params['camera_matrix']
    Rw2c = cam_params['rotation_matrix']
    Tw2c = cam_params['translation_vector']

    #screen_r_vector = data_screen['screen_rotation_vector']
    #print('screen_r_vector = ', screen_r_vector)
    #screen_r_vector[0] = screen_r_vector[0] + 0.18
    
    # Rv2c = cv2.Rodrigues(np.array(screen_r_vector).reshape(3,1))[0]
    # print('Rv2c = ', Rv2c)
    # print('Tv2c = ', Tv2c)
    # Tv2c[0] = Tv2c[0] + x_shift
    # Tv2c[1] = Tv2c[1] + y_shift
    # Tv2c[2] = Tv2c[2] + z_shift

    # Rv2c = [[ 0.96953516 , 0.00218731, -0.24494243],
    #     [-0.0080033 ,  0.99970912 ,-0.02275152],
    #     [ 0.24482142 , 0.02401875,  0.96927064]]
    # Tw2c = [[ -1.84854824],[-92.88120666],[424.69051313]]
   
    # 计算 nc 和 dw2c
    nc = Rw2c[:, 2]
    nc = nc.reshape(-1, 1)
    Tv2c = Tv2c.reshape(-1, 1)
    if nc[2] < 0:
        nc = -nc
    dw2c = np.abs(nc.T @ Tw2c)

    # 计算 Rs2c 和 Ts2c
    Rs2c = np.linalg.inv(np.eye(3) - 2 * np.outer(nc, nc)) @ Rv2c
    assert Tv2c.shape == nc.shape, f"Shape mismatch: Tv2c {Tv2c.shape} != nc {nc.shape}"
    Ts2c = np.linalg.inv(np.eye(3) - 2 * np.outer(nc, nc)) @ (Tv2c - 2 * dw2c * nc)

    # 计算 Rc2w 、Tc2w 、Rs2w 、Ts2w
    Rc2w = np.linalg.inv(Rw2c)
    Tc2w = -Rc2w @ Tw2c

    RTc2w = np.eye(4)
    RTc2w[:3, :3] = Rc2w
    RTc2w[:3, 3] = Tc2w.flatten()

    RTs2c = np.eye(4)
    RTs2c[:3, :3] = Rs2c
    RTs2c[:3, 3] = Ts2c.flatten()
    
    RTs2w = RTc2w @ RTs2c
    Rs2w = RTs2w[:3, :3]
    Ts2w = RTs2w[:3, 3]

    # 定义要保存的数据
    data_to_save = {
        'screen_to_world_rotation': Rs2w,
        'screen_to_world_translation': Ts2w
    }
    return data_to_save





def validate_corners(corners, image_shape):
    """确保角点在图像有效范围内"""
    h, w = image_shape
    for corner in corners:
        x, y = corner.ravel()
        if not (0 <= x < w and 0 <= y < h):
            print(f"Invalid corner: ({x}, {y}) is out of image bounds.")
            return False
    return True

def show_detect_result(img, detections):
    colors = [(255, 0, 0), (0, 255, 0), (0, 0, 255), (255, 255, 0)]
    for detection in detections:
        corners = detection.corners
        tag_id = detection.tag_id
        center = np.mean(corners, axis=0).astype(int)[0]
        for i, corner in enumerate(corners):
            color = colors[i % len(colors)]
            cv2.circle(img, tuple(int(x) for x in corner[0]), 5, color, -1)
        cv2.circle(img, center, 5, (255, 255, 0), -1)
        cv2.putText(img, str(tag_id), (center[0]-40, center[1]), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
        cv2.polylines(img, [np.int32(corners)], isClosed=True, color=(255, 255, 0), thickness=2)

    cv2.namedWindow('AprilGrid Detection', 0)
    cv2.imshow('AprilGrid Detection', img)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


def show_detect_result_new(img, detections):
    colors = [(255, 0, 0), (0, 255, 0), (0, 0, 255), (255, 255, 0)]
    for detection in detections:
        corners = detection.corners
        
        tag_id = detection.tag_id
        print(tag_id)
        print(corners)
        center = np.mean(corners, axis=0).astype(int)
        for i, corner in enumerate(corners):
            color = colors[i % len(colors)]
            cv2.circle(img, tuple(int(x) for x in corner), 5, color, -1)
        cv2.circle(img, center, 5, (255, 255, 0), -1)
        cv2.putText(img, str(tag_id), (center[0]-40, center[1]), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
        cv2.polylines(img, [np.int32(corners)], isClosed=True, color=(255, 255, 0), thickness=2)

    cv2.namedWindow('AprilGrid Detection', 0)
    cv2.imshow('AprilGrid Detection', img)
    cv2.waitKey(100)
    

# 生成 AprilGrid 标定板的 3D 物体坐标，并考虑标签之间的间隙
def generate_3d_points(grid_size, tag_size, tag_spacing):
    """
    grid_size: (rows, cols) 标定板上标签的行数和列数
    tag_size: 每个标签的物理大小（单位:mm)
    tag_spacing: 标签之间的间隙（单位:mm),表示两个标签之间的间隔
    """
    obj_points = []
    
    # 遍历标定板的每一行和每一列
    for i in range(grid_size[0]):  # 行
        for j in range(grid_size[1]):  # 列
            # 每个标签的四个角点坐标
            # 注意：在计算标签位置时，要包括标签之间的间隙
            x_offset = j * (tag_size + tag_spacing)
            y_offset = i * (tag_size + tag_spacing)
            
            obj_points.append([
                [x_offset, y_offset, 0],                         # 左上角
                [x_offset + tag_size, y_offset, 0],              # 右上角
                [x_offset + tag_size, y_offset + tag_size, 0],   # 右下角
                [x_offset, y_offset + tag_size, 0]               # 左下角
            ])
    
    # 将生成的3D点转换为 NumPy 数组并返回
    return np.array(obj_points, dtype=np.float32)


import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def generate_3d_points(grid_size, tag_size, tag_spacing):
    """生成标定板上每个标签的四个角点的3D坐标"""
    obj_points = []
    for i in range(grid_size[0]):  # 行
        for j in range(grid_size[1]):  # 列
            x_offset = j * (tag_size + tag_spacing)
            y_offset = i * (tag_size + tag_spacing)
            obj_points.append([
                [x_offset, y_offset, 0],  # 左上角
                [x_offset + tag_size, y_offset, 0],  # 右上角
                [x_offset + tag_size, y_offset + tag_size, 0],  # 右下角
                [x_offset, y_offset + tag_size, 0]  # 左下角
            ])
    return np.array(obj_points, dtype=np.float32)

def visualize_3d_points(obj_points):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    for idx, tag in enumerate(obj_points):
        xs = [point[0] for point in tag]
        ys = [point[1] for point in tag]
        zs = [point[2] for point in tag]
        ax.plot(xs + [xs[0]], ys + [ys[0]], zs + [zs[0]], marker='o')
        
        # 计算中心点
        center_x = np.mean(xs)
        center_y = np.mean(ys)
        center_z = np.mean(zs)
        
        # 标注ID
        ax.text(center_x, center_y, center_z, str(idx), color='red')

    ax.set_xlabel('X axis')
    ax.set_ylabel('Y axis')
    ax.set_zlabel('Z axis')
    plt.show()
